from django.core.validators import MinValueValidator
from django.db import models
from edc_constants.choices import YES_NO
from edc_constants.constants import NOT_APPLICABLE
from edc_model.models import OtherCharField, date_not_future
from edc_protocol.validators import date_not_before_study_start

from ..choices import BACTERIA_TYPE, BLOOD_CULTURE_RESULTS_ORGANISM, CULTURE_RESULTS


class BloodCultureModelMixin(models.Model):

    blood_culture_performed = models.CharField(max_length=5, choices=YES_NO)

    blood_culture_results = models.CharField(
        verbose_name="Blood culture results, if completed",
        max_length=10,
        choices=CULTURE_RESULTS,
        default=NOT_APPLICABLE,
    )

    blood_taken_date = models.DateField(
        validators=[date_not_before_study_start, date_not_future], null=True, blank=True
    )

    day_blood_taken = models.IntegerField(
        verbose_name="If positive, study day positive culture sample taken",
        validators=[MinValueValidator(1)],
        null=True,
        blank=True,
    )

    blood_culture_organism = models.CharField(
        verbose_name="If growth positive, organism",
        max_length=50,
        choices=BLOOD_CULTURE_RESULTS_ORGANISM,
        default=NOT_APPLICABLE,
    )

    blood_culture_organism_other = OtherCharField(max_length=50, null=True, blank=True)

    bacteria_identified = models.CharField(
        verbose_name="If bacteria, type",
        max_length=50,
        choices=BACTERIA_TYPE,
        default=NOT_APPLICABLE,
    )

    bacteria_identified_other = OtherCharField(max_length=100, null=True, blank=True)

    class Meta:
        abstract = True



from django.db import models
from edc_constants.choices import POS_NEG_NA, YES_NO, YES_NO_NA
from edc_constants.constants import NOT_APPLICABLE
from edc_microbiology.choices import SPUTUM_GENEXPERT
from edc_model.models import date_not_future
from edc_protocol.validators import date_not_before_study_start


class SputumModelMixin(models.Model):

    sputum_afb_performed = models.CharField(
        verbose_name="AFB microscopy performed?",
        max_length=5,
        choices=YES_NO,
        help_text="Was sputum AFB done?",
    )

    sputum_afb_date = models.DateField(
        validators=[date_not_before_study_start, date_not_future], null=True, blank=True
    )

    sputum_results_afb = models.CharField(
        verbose_name="AFB results",
        max_length=10,
        choices=POS_NEG_NA,
        default=NOT_APPLICABLE,
    )

    sputum_performed = models.CharField(
        verbose_name="Culture performed?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    sputum_taken_date = models.DateField(
        validators=[date_not_before_study_start, date_not_future], null=True, blank=True
    )

    sputum_results_culture = models.CharField(
        verbose_name="Culture results",
        max_length=10,
        choices=POS_NEG_NA,
        default=NOT_APPLICABLE,
    )

    sputum_results_positive = models.CharField(
        verbose_name="If culture is positive, please specify:",
        max_length=50,
        null=True,
        blank=True,
    )

    sputum_genexpert_performed = models.CharField(
        verbose_name="Sputum Gene-Xpert performed?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    sputum_genexpert_date = models.DateField(
        verbose_name="Date sputum Gene-Xpert taken",
        validators=[date_not_before_study_start, date_not_future],
        null=True,
        blank=True,
    )

    sputum_result_genexpert = models.CharField(
        verbose_name="Sputum Gene-Xpert results",
        max_length=45,
        choices=SPUTUM_GENEXPERT,
        default=NOT_APPLICABLE,
    )

    class Meta:
        abstract = True



from django.core.validators import MinValueValidator
from django.db import models
from edc_constants.choices import YES_NO
from edc_constants.constants import NOT_APPLICABLE
from edc_model.models import OtherCharField, date_not_future
from edc_protocol.validators import date_not_before_study_start

from ..choices import BIOPSY_RESULTS_ORGANISM, CULTURE_RESULTS


class HistopathologyModelMixin(models.Model):

    tissue_biopsy_taken = models.CharField(max_length=5, choices=YES_NO)

    tissue_biopsy_results = models.CharField(
        verbose_name="If YES, results",
        max_length=10,
        choices=CULTURE_RESULTS,
        default=NOT_APPLICABLE,
    )

    biopsy_date = models.DateField(
        validators=[date_not_before_study_start, date_not_future], null=True, blank=True
    )

    day_biopsy_taken = models.IntegerField(
        verbose_name="If positive, Study day positive culture sample taken",
        validators=[MinValueValidator(1)],
        null=True,
        blank=True,
    )

    tissue_biopsy_organism = models.CharField(
        verbose_name="If growth positive, organism",
        max_length=50,
        choices=BIOPSY_RESULTS_ORGANISM,
        default=NOT_APPLICABLE,
    )

    tissue_biopsy_organism_other = OtherCharField(max_length=50, null=True, blank=True)

    histopathology_report = models.TextField(null=True, blank=True)

    class Meta:
        abstract = True



from edc_model import models as edc_models

from .model_mixins import (
    BloodCultureModelMixin,
    CsfModelMixin,
    HistopathologyModelMixin,
    SputumModelMixin,
    UrineCultureModelMixin,
)


class Microbiology(
    UrineCultureModelMixin,
    BloodCultureModelMixin,
    SputumModelMixin,
    CsfModelMixin,
    HistopathologyModelMixin,
    edc_models.BaseUuidModel,
):
    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Microbiology"
        verbose_name_plural = "Microbiology"



def get_urine_culture_fieldset():
    return [
        "Urine Culture (Only for patients with >50 white cells in urine)",
        {
            "fields": (
                "urine_culture_performed",
                "urine_taken_date",
                "urine_culture_results",
                "urine_culture_organism",
                "urine_culture_organism_other",
            )
        },
    ]


def get_blood_culture_fieldset():
    return [
        "Blood Culture",
        {
            "fields": (
                "blood_culture_performed",
                "blood_culture_results",
                "blood_taken_date",
                "day_blood_taken",
                "blood_culture_organism",
                "blood_culture_organism_other",
                "bacteria_identified",
                "bacteria_identified_other",
            )
        },
    ]


def get_sputum_fieldset():
    return [
        "Sputum Microbiology",
        {
            "fields": (
                "sputum_afb_performed",
                "sputum_afb_date",
                "sputum_results_afb",
                "sputum_performed",
                "sputum_taken_date",
                "sputum_results_culture",
                "sputum_results_positive",
                "sputum_genexpert_performed",
                "sputum_genexpert_date",
                "sputum_result_genexpert",
            )
        },
    ]


def get_csf_fieldset():
    return [
        "CSF Microbiology",
        {
            "fields": (
                "csf_genexpert_performed",
                "csf_genexpert_date",
                "csf_result_genexpert",
            )
        },
    ]


def get_histopathology_fieldset():
    return [
        "Histopathology",
        {
            "fields": (
                "tissue_biopsy_taken",
                "tissue_biopsy_results",
                "biopsy_date",
                "day_biopsy_taken",
                "tissue_biopsy_organism",
                "tissue_biopsy_organism_other",
                "histopathology_report",
            )
        },
    ]



from django import forms
from edc_form_validators import FormValidatorMixin
from edc_sites.forms import SiteModelFormMixin

from .form_validators import MicrobiologyFormValidator
from .models import Microbiology


class MicrobiologyForm(SiteModelFormMixin, FormValidatorMixin, forms.ModelForm):

    form_validator_cls = MicrobiologyFormValidator

    class Meta:
        model = Microbiology
        fields = "__all__"



from edc_constants.constants import NOT_APPLICABLE, OTHER, POS

from .constants import (
    BACTERIA,
    CRYPTOCOCCUS_NEOFORMANS,
    ECOLI,
    KLEBSIELLA_SPP,
    NO_GROWTH,
)

BACTERIA_TYPE = (
    (NOT_APPLICABLE, "Not applicable"),
    (ECOLI, "E.coli"),
    (KLEBSIELLA_SPP, "Klebsiella spp."),
    ("streptococcus_pneumoniae", "Streptococcus pneumoniae"),
    ("staphylococus_aureus", "(Sensitive) Staphylococus aureus"),
    ("mrsa", "MRSA"),
    (OTHER, "Other"),
)

BLOOD_CULTURE_RESULTS_ORGANISM = (
    (NOT_APPLICABLE, "Not applicable"),
    (CRYPTOCOCCUS_NEOFORMANS, "Cryptococcus neoformans"),
    (BACTERIA, "Bacteria"),
    ("bacteria_and_cryptococcus", "Bacteria and Cryptococcus"),
    (OTHER, "Other"),
)

BIOPSY_RESULTS_ORGANISM = (
    (NOT_APPLICABLE, "Not applicable"),
    (CRYPTOCOCCUS_NEOFORMANS, "Cryptococcus neoformans"),
    ("mycobacterium_tuberculosis", "Mycobacterium Tuberculosis"),
    (OTHER, "Other"),
)

CULTURE_RESULTS = (
    (NOT_APPLICABLE, "Not applicable"),
    (NO_GROWTH, "No growth"),
    (POS, "Positive"),
)

URINE_CULTURE_RESULTS_ORGANISM = (
    (NOT_APPLICABLE, "Not applicable"),
    (ECOLI, "E.coli"),
    (KLEBSIELLA_SPP, "Klebsiella spp."),
    (OTHER, "Other"),
)

SPUTUM_GENEXPERT = (
    ("mtb_detected_rif_resistance_detected", "MTB DETECTED & Rif Resistance DETECTED"),
    (
        "mtb_detected_rif_resistance_not_detected",
        "MTB DETECTED & Rif Resistance NOT detected",
    ),
    (
        "mtb_detected_rif_resistance_indeterminate",
        "MTB DETECTED & Rif Resistance INDETERMINATE",
    ),
    ("mtb_not_detected", "MTB NOT detected"),
    (NOT_APPLICABLE, "Not applicable"),
)



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple

from .fieldsets import (
    get_blood_culture_fieldset,
    get_csf_fieldset,
    get_histopathology_fieldset,
    get_sputum_fieldset,
    get_urine_culture_fieldset,
)


class MicrobiologyModelAdminMixin:

    fieldsets = (
        get_urine_culture_fieldset(),
        get_blood_culture_fieldset(),
        get_sputum_fieldset(),
        get_csf_fieldset(),
        get_histopathology_fieldset(),
        audit_fieldset_tuple,
    )

    radio_fields = {
        "urine_culture_performed": admin.VERTICAL,
        "urine_culture_results": admin.VERTICAL,
        "urine_culture_organism": admin.VERTICAL,
        "blood_culture_performed": admin.VERTICAL,
        "blood_culture_results": admin.VERTICAL,
        "blood_culture_organism": admin.VERTICAL,
        "bacteria_identified": admin.VERTICAL,
        "sputum_afb_performed": admin.VERTICAL,
        "sputum_results_afb": admin.VERTICAL,
        "sputum_performed": admin.VERTICAL,
        "sputum_results_culture": admin.VERTICAL,
        "sputum_result_genexpert": admin.VERTICAL,
        "sputum_genexpert_performed": admin.VERTICAL,
        "csf_result_genexpert": admin.VERTICAL,
        "csf_genexpert_performed": admin.VERTICAL,
        "tissue_biopsy_taken": admin.VERTICAL,
        "tissue_biopsy_results": admin.VERTICAL,
        "tissue_biopsy_organism": admin.VERTICAL,
    }



from django.db import models
from edc_constants.choices import YES_NO
from edc_constants.constants import NOT_APPLICABLE
from edc_model.models import OtherCharField, date_not_future
from edc_protocol.validators import date_not_before_study_start

from ..choices import CULTURE_RESULTS, URINE_CULTURE_RESULTS_ORGANISM


class UrineCultureModelMixin(models.Model):

    urine_culture_performed = models.CharField(
        max_length=5,
        choices=YES_NO,
        help_text="only for patients with >50 white cells in urine",
    )

    urine_taken_date = models.DateField(
        validators=[date_not_before_study_start, date_not_future], null=True, blank=True
    )

    urine_culture_results = models.CharField(
        verbose_name="Urine culture results, if completed",
        max_length=10,
        choices=CULTURE_RESULTS,
        default=NOT_APPLICABLE,
    )

    urine_culture_organism = models.CharField(
        verbose_name="If positive, organism",
        max_length=25,
        choices=URINE_CULTURE_RESULTS_ORGANISM,
        default=NOT_APPLICABLE,
    )

    urine_culture_organism_other = OtherCharField(max_length=50, null=True, blank=True)

    class Meta:
        abstract = True



from edc_constants.constants import OTHER, POS, YES
from edc_form_validators import FormValidator
from edc_form_validators.extra_mixins import StudyDayFormValidatorMixin

from .constants import BACTERIA


class MicrobiologyFormValidator(StudyDayFormValidatorMixin, FormValidator):
    def clean(self):

        self.validate_study_day_with_datetime(
            subject_identifier=self.cleaned_data.get("subject_visit").subject_identifier,
            study_day=self.cleaned_data.get("day_blood_taken"),
            compare_date=self.cleaned_data.get("blood_taken_date"),
            study_day_field="day_blood_taken",
        )

        self.validate_study_day_with_datetime(
            subject_identifier=self.cleaned_data.get("subject_visit").subject_identifier,
            study_day=self.cleaned_data.get("day_biopsy_taken"),
            compare_date=self.cleaned_data.get("biopsy_date"),
            study_day_field="day_biopsy_taken",
        )

        self.required_if(
            YES, field="urine_culture_performed", field_required="urine_taken_date"
        )

        self.applicable_if(
            YES,
            field="urine_culture_performed",
            field_applicable="urine_culture_results",
        )

        self.applicable_if(
            POS,
            field="urine_culture_results",
            field_applicable="urine_culture_organism",
        )

        self.validate_other_specify(
            field="urine_culture_organism",
            other_specify_field="urine_culture_organism_other",
            other_stored_value=OTHER,
        )

        self.applicable_if(
            YES,
            field="blood_culture_performed",
            field_applicable="blood_culture_results",
        )

        self.required_if(
            YES, field="blood_culture_performed", field_required="blood_taken_date"
        )

        self.required_if(POS, field="blood_culture_results", field_required="day_blood_taken")

        self.applicable_if(
            POS,
            field="blood_culture_results",
            field_applicable="blood_culture_organism",
        )

        self.validate_other_specify(
            field="blood_culture_organism",
            other_specify_field="blood_culture_organism_other",
            other_stored_value=OTHER,
        )

        condition = (
            self.cleaned_data.get("blood_culture_organism") == BACTERIA
            or self.cleaned_data.get("blood_culture_organism") == "bacteria_and_cryptococcus"
        )
        self.applicable_if_true(condition=condition, field_applicable="bacteria_identified")

        self.validate_other_specify(
            field="bacteria_identified",
            other_specify_field="bacteria_identified_other",
            other_stored_value=OTHER,
        )

        self.required_if(YES, field="sputum_afb_performed", field_required="sputum_afb_date")

        self.applicable_if(
            YES, field="sputum_afb_performed", field_applicable="sputum_results_afb"
        )

        self.required_if(YES, field="sputum_performed", field_required="sputum_taken_date")

        self.applicable_if(
            YES, field="sputum_performed", field_applicable="sputum_results_culture"
        )

        self.required_if(
            POS,
            field="sputum_results_culture",
            field_required="sputum_results_positive",
        )

        self.required_if(
            YES,
            field="sputum_genexpert_performed",
            field_required="sputum_genexpert_date",
        )

        self.applicable_if(
            YES,
            field="sputum_genexpert_performed",
            field_applicable="sputum_result_genexpert",
        )

        self.required_if(
            YES,
            field="csf_genexpert_performed",
            field_required="csf_genexpert_date",
        )

        self.applicable_if(
            YES,
            field="csf_genexpert_performed",
            field_applicable="csf_result_genexpert",
        )

        self.applicable_if(
            YES,
            field="tissue_biopsy_taken",
            field_applicable="tissue_biopsy_results",
        )

        self.required_if(YES, field="tissue_biopsy_taken", field_required="biopsy_date")

        self.required_if(POS, field="tissue_biopsy_results", field_required="day_biopsy_taken")

        self.applicable_if(
            POS,
            field="tissue_biopsy_results",
            field_applicable="tissue_biopsy_organism",
        )

        self.validate_other_specify(
            field="tissue_biopsy_organism",
            other_specify_field="tissue_biopsy_organism_other",
            other_stored_value=OTHER,
        )



from .blood_culture import BloodCultureModelMixin
from .csf import CsfModelMixin
from .histopathology import HistopathologyModelMixin
from .sputum import SputumModelMixin
from .urine_culture import UrineCultureModelMixin



from django.apps import apps as django_apps

EDC_MICROBIOLOGY = "EDC_MICROBIOLOGY"
EDC_MICROBIOLOGY_SUPER = "EDC_MICROBIOLOGY_SUPER"
EDC_MICROBIOLOGY_VIEW = "EDC_MICROBIOLOGY_VIEW"

codenames = []
app_config = django_apps.get_app_config("edc_microbiology")
for model_cls in app_config.get_models():
    if "historical" not in model_cls._meta.label_lower:
        for action in ["view_", "add_", "change_", "delete_", "view_historical"]:
            codenames.append(f".{action}".join(model_cls._meta.label_lower.split(".")))
codenames.sort()
